/* -*- Mode: C++; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* vim: set ts=8 sts=2 et sw=2 tw=80: */
/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

#ifndef mozilla_ServoStyleContext_h
#define mozilla_ServoStyleContext_h

#include "nsIMemoryReporter.h"
#include "nsStyleContext.h"
#include "nsWindowSizes.h"
#include <algorithm>

namespace mozilla {

namespace dom {
class Element;
} // namespace dom

MOZ_DEFINE_MALLOC_SIZE_OF(ServoComputedValuesMallocSizeOf)

class ServoStyleContext final : public nsStyleContext
{
public:
  ServoStyleContext(nsPresContext* aPresContext,
                    nsIAtom* aPseudoTag,
                    CSSPseudoElementType aPseudoType,
                    ServoComputedDataForgotten aComputedValues);

  nsPresContext* PresContext() const { return mPresContext; }
  const ServoComputedData* ComputedData() const { return &mSource; }

  void AddRef() { Servo_StyleContext_AddRef(this); }
  void Release() { Servo_StyleContext_Release(this); }

  ServoStyleContext* GetStyleIfVisited() const
  {
    return ComputedData()->visited_style.mPtr;
  }

  ServoStyleContext* GetCachedInheritingAnonBoxStyle(nsIAtom* aAnonBox) const;

  void SetCachedInheritedAnonBoxStyle(nsIAtom* aAnonBox,
                                      ServoStyleContext& aStyle)
  {
    MOZ_ASSERT(!GetCachedInheritingAnonBoxStyle(aAnonBox));
    MOZ_ASSERT(!aStyle.mNextInheritingAnonBoxStyle);

    // NOTE(emilio): Since we use it to cache inheriting anon boxes in a linked
    // list, we can't use that cache if the style we're inheriting from is an
    // inheriting anon box itself, since otherwise our parent would mistakenly
    // think that the style we're caching inherits from it.
    //
    // See the documentation of mNextInheritingAnonBoxStyle.
    if (IsInheritingAnonBox()) {
      return;
    }

    mNextInheritingAnonBoxStyle.swap(aStyle.mNextInheritingAnonBoxStyle);
    mNextInheritingAnonBoxStyle = &aStyle;
  }

  /**
   * Makes this context match |aOther| in terms of which style structs have
   * been resolved.
   */
  inline void ResolveSameStructsAs(const ServoStyleContext* aOther);

  void AddSizeOfIncludingThis(SizeOfState& aState, nsStyleSizes& aSizes,
                              bool aIsDOM) const
  {
    // XXX WARNING: similar to ServoComputedData::AddSizeOfExcludingThis(),
    // but here we need to step back 4 or 8 bytes to get past the servo_arc::Arc
    // refcount to the base pointer.
    static_assert(alignof(ServoStyleContext) == 4 ||
                  alignof(ServoStyleContext) == 8,
                  "alignment will break AddSizeOfExcludingThis()");
    const char* p = reinterpret_cast<const char*>(this);
    p -= std::max(sizeof(size_t), alignof(ServoStyleContext));

    // We use ServoComputedValuesMallocSizeOf rather than
    // |aState.mMallocSizeOf| to better distinguish in DMD's output the memory
    // measured here.
    if (aIsDOM) {
      aSizes.mComputedValuesDom += ServoComputedValuesMallocSizeOf(p);
    } else {
      aSizes.mComputedValuesNonDom += ServoComputedValuesMallocSizeOf(p);
    }
    mSource.AddSizeOfExcludingThis(aState, aSizes);
  }

private:
  nsPresContext* mPresContext;
  ServoComputedData mSource;

  // A linked-list cache of inheriting anon boxes inheriting from this style _if
  // the style isn't an inheriting anon-box_.
  //
  // Otherwise it represents the next entry in the cache of the parent style
  // context.
  RefPtr<ServoStyleContext> mNextInheritingAnonBoxStyle;
};

} // namespace mozilla

#endif // mozilla_ServoStyleContext_h
